﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Threading;
using HEAL.Attic;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Regression {
  [StorableType("4318C6BD-E0A1-45FE-AC30-96E7F73B51FB")]
  public class SymbolicRegressionConstraintAnalyzer : SymbolicDataAnalysisAnalyzer, ISymbolicExpressionTreeAnalyzer {
    private const string ConstraintViolationsResultName = "Constraint Violations";
    private const string ProblemDataParameterName = "ProblemData";
    private const string ConstraintViolationParameterName = "ConstraintViolations";
    private const string PenalityMultiplierParameterName = "ConstraintsPenalityMultiplier";
    private const string PenalityMultiplierResultParameterName = "PenalityMultiplierResultParameter";
    private const string StepSizeParameterName = "Step Size";
    private const string MaximumStepsParameterName = "Maximum Steps";
    private const string StartPenalityParameterName = "Start Penality";

    #region parameter properties

    public ILookupParameter<IRegressionProblemData> RegressionProblemDataParameter =>
      (ILookupParameter<IRegressionProblemData>) Parameters[ProblemDataParameterName];

    public IResultParameter<DataTable> ConstraintViolationParameter =>
      (IResultParameter<DataTable>) Parameters[ConstraintViolationParameterName];

    public ILookupParameter<DoubleValue> PenalityMultiplierParameter =>
      (ILookupParameter<DoubleValue>) Parameters[PenalityMultiplierParameterName];

    public IValueParameter<IntValue> StepSizeParameter =>
      (IValueParameter<IntValue>) Parameters[StepSizeParameterName];

    public IValueParameter<IntValue> MaximumStepsParameter =>
      (IValueParameter<IntValue>) Parameters[MaximumStepsParameterName];

    public IResultParameter<DataTable> PenalityMultiplierResultParameter =>
      (IResultParameter<DataTable>) Parameters[PenalityMultiplierResultParameterName];

    public IValueParameter<DoubleValue> StartPenalityParameter =>
      (IValueParameter<DoubleValue>) Parameters[StartPenalityParameterName];

    #endregion

    public override bool EnabledByDefault => false;
    public static int Iterations { get; set; } = 0;

    [StorableConstructor]
    protected SymbolicRegressionConstraintAnalyzer(StorableConstructorFlag _) : base(_) { }

    protected SymbolicRegressionConstraintAnalyzer(SymbolicRegressionConstraintAnalyzer original, Cloner cloner) :
      base(original, cloner) { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicRegressionConstraintAnalyzer(this, cloner);
    }

    public SymbolicRegressionConstraintAnalyzer() {
      Parameters.Add(new LookupParameter<IRegressionProblemData>(ProblemDataParameterName,
        "The problem data of the symbolic data analysis problem."));
      Parameters.Add(new ResultParameter<DataTable>(ConstraintViolationParameterName,
        "Shows the number of constraint violations!"));
      Parameters.Add(new LookupParameter<DoubleValue>(PenalityMultiplierParameterName,
        "Lookup parameter for the penality multiplier.", "PenalityMultiplier"));
      Parameters.Add(new ResultParameter<DataTable>(PenalityMultiplierResultParameterName,
        "Shows the behavior of the penality multiplier."));
      Parameters.Add(new ValueParameter<IntValue>(StepSizeParameterName,
        "Defines the step size for the increasing penality multiplier.", new IntValue(1)));
      Parameters.Add(new ValueParameter<IntValue>(MaximumStepsParameterName,
        "Defines maximum steps for the increasing penality multiplier.", new IntValue(1000)));
      Parameters.Add(new ValueParameter<DoubleValue>(StartPenalityParameterName,
        "The start value for the penality multiplier.", new DoubleValue(0.5)));


      ConstraintViolationParameter.DefaultValue = new DataTable(ConstraintViolationParameterName) {
        VisualProperties = {
          XAxisTitle = "Generations",
          YAxisTitle = "Constraint Violations"
        }
      };

      PenalityMultiplierResultParameter.DefaultValue = new DataTable(PenalityMultiplierResultParameterName) {
        VisualProperties = {
          XAxisTitle = "Generations",
          YAxisTitle = "Penality Multiplier"
        }
      };
    }

    public override void InitializeState() {
      Iterations = 0;
      base.InitializeState();
    }

    public override void ClearState() {
      Iterations = 0;
      base.ClearState();
    }


    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (!Parameters.ContainsKey(PenalityMultiplierParameterName))
        Parameters.Add(new LookupParameter<DoubleValue>(PenalityMultiplierParameterName,
          "Lookup parameter for the penality multiplier.", "PenalityMultiplier"));

      if (!Parameters.ContainsKey(PenalityMultiplierResultParameterName)) {
        Parameters.Add(new ResultParameter<DataTable>(PenalityMultiplierResultParameterName,
          "Shows the behavior of the penality multiplier."));

        PenalityMultiplierResultParameter.DefaultValue = new DataTable(PenalityMultiplierResultParameterName) {
          VisualProperties = {
            XAxisTitle = "Generations",
            YAxisTitle = "Penality Multiplier"
          }
        };
      }

      if (!Parameters.ContainsKey(StepSizeParameterName))
        Parameters.Add(new ValueParameter<IntValue>(StepSizeParameterName,
          "Defines the step size for the increasing penality multiplier.", new IntValue(1)));

      if (!Parameters.ContainsKey(MaximumStepsParameterName))
        Parameters.Add(new ValueParameter<IntValue>(MaximumStepsParameterName,
          "Defines maximum steps for the increasing penality multiplier.", new IntValue(1000)));


      if (!Parameters.ContainsKey(StartPenalityParameterName))
        Parameters.Add(new ValueParameter<DoubleValue>(StartPenalityParameterName,
          "The start value for the penality multiplier.", new DoubleValue(0.5)));
    }

    public override IOperation Apply() {
      Iterations++;
      var rowName = "GeneralizedExponentialDiscreteDoubleValueModifier";
      var stepSize = StepSizeParameter?.Value?.Value ?? 1;
      var startValue = StartPenalityParameter?.Value?.Value ?? 0.5;
      var maximumSteps = MaximumStepsParameter?.Value?.Value ?? 1000;
      var penalityMultiplier = new DoubleValue(LinearDiscreteDoubleValueModifier.Apply(0, startValue, 1.0,
        (int) (Iterations / stepSize) * stepSize, 0, maximumSteps));
      PenalityMultiplierParameter.ActualValue = penalityMultiplier;
      //GeneralizedExponentialDiscreteDoubleValueModifier
      //QuadraticDiscreteDoubleValueModifier
      //LinearDiscreteDoubleValueModifier

      var problemData = RegressionProblemDataParameter.ActualValue;
      var trees = SymbolicExpressionTreeParameter.ActualValue;

      var results = ResultCollectionParameter.ActualValue;
      var constraints = problemData.IntervalConstraints.EnabledConstraints;
      var variableRanges = problemData.VariableRanges.GetReadonlyDictionary();
      var newDataTable = ConstraintViolationParameter.ActualValue;

      if (newDataTable.Rows.Count == 0)
        foreach (var constraint in constraints)
          newDataTable.Rows.Add(new DataRow(constraint.Expression));

      var interpreter = new IntervalInterpreter();
      foreach (var constraint in constraints) {
        var violations = trees.Select(tree =>
                                 ConstraintSatisfied(constraint, interpreter, variableRanges, tree, out double error))
                              .Count(satisfied => !satisfied);
        newDataTable.Rows[constraint.Expression].Values.Add(violations);
      }

      var penalityDataTable = PenalityMultiplierResultParameter.ActualValue;
      if (penalityDataTable.Rows.Count == 0)
        penalityDataTable.Rows.Add(new DataRow(rowName));
      penalityDataTable.Rows[rowName].Values.Add(penalityMultiplier.Value);

      return base.Apply();
    }

    public static bool ConstraintSatisfied(IntervalConstraint constraint,
                                           IntervalInterpreter intervalInterpreter,
                                           IReadOnlyDictionary<string, Interval> variableRanges,
                                           ISymbolicExpressionTree solution,
                                           out double error) {
      if (constraint.Variable != null && !variableRanges.ContainsKey(constraint.Variable))
        throw new
          ArgumentException($"The given variable {constraint.Variable} in the constraint does not exists in the model.",
            nameof(IntervalConstraintsParser));

      Interval resultInterval;

      //Create new variable ranges for defined regions
      var regionRanges = new IntervalCollection();
      foreach (var kvp in variableRanges) {
        if (kvp.Key != constraint.Target &&
            constraint.Regions.GetReadonlyDictionary().TryGetValue(kvp.Key, out Interval val)) {
          regionRanges.AddInterval(kvp.Key, val);
        } else {
          regionRanges.AddInterval(kvp.Key, kvp.Value);
        }
      }

      // calculate result interval
      if (!constraint.IsDerivative) {
        resultInterval =
          intervalInterpreter.GetSymbolicExpressionTreeInterval(solution,
            regionRanges.GetReadonlyDictionary());
      } else {
        var tree = solution;
        for (var i = 0; i < constraint.NumberOfDerivations; ++i) {
          if (!IntervalInterpreter.IsCompatible(tree) || !DerivativeCalculator.IsCompatible(tree))
            throw new ArgumentException("Cube, Root, Power Symbols are not supported.");
          tree = DerivativeCalculator.Derive(tree, constraint.Variable);
        }

        resultInterval =
          intervalInterpreter.GetSymbolicExpressionTreeInterval(tree,
            regionRanges.GetReadonlyDictionary());
      }

      error = 0.0;
      bool satisfied = true;
      if (!constraint.Interval.Contains(resultInterval.LowerBound)) {
        satisfied = false;
        error += Math.Abs(resultInterval.LowerBound - constraint.Interval.LowerBound);
      }

      if (!constraint.Interval.Contains(resultInterval.UpperBound)) {
        satisfied = false;
        error += Math.Abs(resultInterval.UpperBound - constraint.Interval.UpperBound);
      }

      error *= constraint.Weight;

      //var satisfied = constraint.Interval.Contains(resultInterval);
      return satisfied;
    }

    public static bool ConstraintsSatisfied(IEnumerable<IntervalConstraint> constraints,
                                            IReadOnlyDictionary<string, Interval> variableRanges,
                                            ISymbolicExpressionTree solution,
                                            out double error) {
      var intervalInterpreter = new IntervalInterpreter();
      error = 0.0;
      var satisfied = true;
      foreach (var constraint in constraints) {
        if (constraint.Variable != null && !variableRanges.ContainsKey(constraint.Variable))
          throw new
            ArgumentException(
              $"The given variable {constraint.Variable} in the constraint does not exists in the model.",
              nameof(IntervalConstraintsParser));

        satisfied = ConstraintSatisfied(constraint, intervalInterpreter, variableRanges, solution, out double e) &&
                    satisfied;
        error += e;
        //if (!satisfied) return false;
      }

      return satisfied;
    }
  }
}